from abc import ABC, abstractmethod

from torch import nn


class Logic(ABC):
    """
    Implements an abstract logic
    """
    @abstractmethod
    class Negation(nn.Module):
        """
        Torch module that implements the Negation Network
        Must be implemented by AbstractLogic subclasses
        """
        pass

    @abstractmethod
    class TNorm(nn.Module):
        """
        Torch module that implements the TNorm Network
        Must be implemented by AbstractLogic subclasses
        """
        pass

    @abstractmethod
    class TConorm(nn.Module):
        """
        Torch module that implements the TConorm Network
        Must be implemented by AbstractLogic subclasses
        """
        pass

    @abstractmethod
    class Equal(nn.Module):
        """
        Torch module that implements the Equal Network
        Must be implemented by AbstractLogic subclasses
        """
        pass

    @abstractmethod
    class Residual(nn.Module):
        """
        Torch module that implements the Residual Network
        Must be implemented by AbstractLogic subclasses
        """
        pass

    @property
    def negation(self) -> nn.Module:
        """
        :return: a new instance of the Negation network.
        :rtype: nn.Module
        """
        assert issubclass(self.Negation, nn.Module)
        return self.Negation()

    @property
    def t_norm(self) -> nn.Module:
        """
        :return: a new instance of the TNorm network.
        :rtype: nn.Module
        """
        assert issubclass(self.TNorm, nn.Module)
        return self.TNorm()

    @property
    def t_conorm(self) -> nn.Module:
        """
        :return: a new instance of the TConorm network.
        :rtype: nn.Module
        """
        assert issubclass(self.TConorm, nn.Module)
        return self.TConorm()

    @property
    def equal(self) -> nn.Module:
        """
        :return: a new instance of the Equal network.
        :rtype: nn.Module
        """
        assert issubclass(self.Equal, nn.Module)
        return self.Equal()

    @property
    def residual(self) -> nn.Module:
        """
        :return: a new instance of the Residual network.
        :rtype: nn.Module
        """
        assert issubclass(self.Residual, nn.Module)
        return self.Residual()



import torch.nn.functional as F
from torch import nn

from pymetheus.utils.functionalities import *


class QuantifiedFormula(nn.Module):
    """
    Implements a quantified formula that supports the grouping of the variables to avoid generating inputs that are
    too big to handle
    """

    def __init__(self, parsed_rule, networks, variables, aggregation=lambda x: torch.mean(x)):
        """
        :param parsed_rule: the rule computed by the following network
        :param networks: the networks in LogicNet
        :param variables: the variables to be considered in the current network
        :param aggregation: the aggregation function to be used for the quantification (e.g., the mean)
        """
        super(QuantifiedFormula, self).__init__()
        self.parsed_rule = parsed_rule  # the rule
        self.vars = variables  # vars of the rule (?a, ?b)
        self.aggregator = aggregation  # (lambda x : torch.mean())

        explored = list(set(get_all_networks(parsed_rule)))  # explore the rule to get network ids (e.g., "country")

        # if it is a simple one predicate only axiom (e.g., forall x: smokes(x)) the method returns a string
        nets_name = (list(flatten(explored)))
        to_module_dict = list(map(lambda x: [x, networks[x]], nets_name))

        self.nns = nn.ModuleDict(to_module_dict)

    def compute(self, parsed, vars):
        """
        Recursive function that goes over the tree and computes the truth values
        :param parsed:
        :param vars:
        :return:
        """
        if parsed.value in ["->", "&", "|", "%"]:
            left = self.compute(parsed.children[0], vars)
            right = self.compute(parsed.children[1], vars)

            before = self.nns[parsed.value](left, right)

            return before

        if parsed.value in ["~"]:
            return 1 - self.compute(parsed.children[0], vars)

        if parsed.value in self.nns.keys():  # if the value is a function or a predicate we need to compute it
            accumulate = []
            for v in parsed.children:
                accumulate.append(self.compute(v, vars))

            if not self.nns[parsed.value].system:
                after_run = self.nns[parsed.value](*accumulate)
                return after_run

            val = self.nns[parsed.value](accumulate)
            return val

        if parsed.value[0] == "?":
            return torch.stack(vars[parsed.value])

    def forward(self, variables):  # {"?a" : a, "?b" : b}
        computed = self.compute(self.parsed_rule, variables)
        return computed


class Function(nn.Module):
    def __init__(self, input_size, output_size):
        super(Function, self).__init__()
        self.system = True
        self.linear = nn.Linear(input_size, output_size)

    def ingest(self, ingestion):
        from_map_to_list = list(ingestion)
        get_lists_of_arguments = map(list, zip(*from_map_to_list))
        concatenate_inputs = list(map(torch.cat, get_lists_of_arguments))

        stack_inputs_for_network = torch.stack(concatenate_inputs)

        return stack_inputs_for_network

    def forward(self, x):
        x = self.ingest(x)
        return self.linear(x)


class Predicate(nn.Module):
    def __init__(self, size):
        super(Predicate, self).__init__()
        k = 10
        self.system = True
        self.pos_emb = nn.Embedding(2, size)

        self.W = nn.Bilinear(size, size, k)
        self.V = nn.Linear(size, k)
        self.u = nn.Linear(k, 1)
        self.ul = torch.randn(k, requires_grad=True, device=get_torch_device())

        self.to_output = nn.Linear(size, 1)

        self.tokeys = nn.Linear(size, size, bias=False)
        self.toqueries = nn.Linear(size, size, bias=False)
        self.tovalues = nn.Linear(size, size, bias=False)

    def self_attention(self, ingestion):
        get_lists_of_arguments = list(map(torch.stack, zip(*ingestion)))
        data = (torch.stack(get_lists_of_arguments))

        (b, t, k) = data.size()

        positions = torch.arange(t, device=get_torch_device())
        positions = self.pos_emb(positions)[None, :, :].expand((b, t, k))

        data = positions + data
        queries = self.toqueries(data).view(b, t, k)
        keys = self.tokeys(data).view(b, t, k)
        values = self.tovalues(data).view(b, t, k)

        keys = keys.transpose(1, 2).contiguous().view(b, t, k)
        queries = queries.transpose(1, 2).contiguous().view(b, t, k)

        # queries = queries / (k ** (1/4))
        # keys = keys / (k ** (1/4))

        dot = torch.bmm(queries, keys.transpose(1, 2))

        dot = F.softmax(dot, dim=2)
        out = torch.bmm(dot, values).view(b, t, k)

        out = out.transpose(1, 2).contiguous().view(b, t, k)
        return out.mean(dim=1)

    def attention(self, ingestion):
        get_lists_of_arguments = list(map(torch.stack, zip(*ingestion)))
        data = (torch.stack(get_lists_of_arguments))

        (b, t, k) = data.size()

        positions = torch.arange(t, device=get_torch_device())
        positions = self.pos_emb(positions)[None, :, :].expand((b, t, k))

        data = positions + data

        raw_weights = torch.bmm(data, data.transpose(1, 2))

        weights = F.softmax(raw_weights, dim=2)

        y = torch.bmm(weights, data)
        # v = y.reshape(b, t*k)

        y = y.mean(dim=1)

        return y

    def ingest(self, ingestion):
        from_map_to_list = list(ingestion)
        get_lists_of_arguments = map(list, zip(*from_map_to_list))
        concatenate_inputs = list(map(torch.cat, get_lists_of_arguments))

        stack_inputs_for_network = torch.stack(concatenate_inputs)

        return stack_inputs_for_network

    def forward(self, x):
        x = self.ingest(x)

        first = self.W(x, x)
        second = self.V(x)
        output = torch.relu(first + second)
        x = self.u(output)

        # x = self.to_output(x)
        return torch.sigmoid(x)


class LinearPredicate(Predicate):
    def __init__(self, size):
        super().__init__(size)
        self.linear_one = nn.Linear(size, 100)
        self.linear_two = nn.Linear(100, 50)
        self.linear_three = nn.Linear(50, 1)

    def forward(self, x):
        x = self.ingest(x)

        x = self.linear_one(x)
        x = torch.relu(x)
        x = self.linear_two(x)
        x = torch.relu(x)
        x = self.linear_three(x)

        return torch.sigmoid(x)



from enum import Enum, unique

from pymetheus.logics.interfaces import *
from pymetheus.logics.lukasiewicz_logic import LukasiewiczLogic


@unique
class LogicEnum(Enum):
    """
    Enumerates the logics available for LogicNet
    """
    LUKASIEWICZ = LukasiewiczLogic()

    @classmethod
    def default(cls) -> 'LogicEnum':
        """
        :return: the default logic
        :rtype: LogicEnum
        """
        return cls.LUKASIEWICZ



class DobuleInitalizationException(Exception):
    def __init__(self, message, error):

        message = message + " | Check the following parameter: " + error + "."

        super().__init__(message)



import collections
from functools import lru_cache
from itertools import chain, islice

import torch


def harmonic_mean(input):
    return input.pow(-1).mean().pow(-1)


class Node(object):

    def __init__(self, value, left=None, right=None):
        self.value = value  # The node value
        self.left = left  # Left child
        self.right = right  # Right child


def get_networks_ids(node):
    """
    Gets network ids, might return list in case of multiple predicates or string
    :param node:
    :return:
    """
    accumulator = []

    if node.value in ["->", "&", "|"]:
        accumulator.append(get_networks_ids(node.left))
        accumulator.append(get_networks_ids(node.right))
        accumulator.append(node.value)
        return accumulator
    elif node.left is None and node.right is None:  # leaf
        if node.value[0] == "~":
            network_id = node.value[1][0]
            return network_id
        else:
            network_id = node.value[0]
            return network_id


def batching(n, iterable):
    iterable = iter(iterable)
    while True:
        try:
            yield chain([next(iterable)], islice(iterable, n - 1))
        except StopIteration:
            return


def flatten(l):
    for el in l:
        if isinstance(el, collections.Iterable) and not isinstance(el, (str, bytes)):
            yield from flatten(el)
        else:
            yield el


def exploring(node, accum=""):
    print(accum + str(node.value))
    accum = accum + "\t"
    for a in node.children:
        (exploring(a, accum))


def get_all_networks(node):
    ids = []

    if node.value[0] != "?":
        ids.append(node.value)
        for a in node.children:
            ids.append(get_all_networks(a))

    return flatten(ids)


class MultiNode(object):

    def __init__(self, value, children):
        self.value = value
        self.children = children


def rule_to_tree_augmented(parsed_network, number=0):
    """
    Takes a parsed network and generates a tree structure
    :param parsed_network:
    :return:
    """

    if len(parsed_network) == 3:
        left_tree = rule_to_tree_augmented(parsed_network[0])
        right_tree = rule_to_tree_augmented(parsed_network[2])
        s_node = MultiNode(parsed_network[1], [])

        s_node.children = [left_tree, right_tree]
        return s_node

    if str(parsed_network[0]) == "~":
        s_node = MultiNode(parsed_network[0], [])
        get_value = rule_to_tree_augmented(parsed_network[1], number=number + 1)
        s_node.children.append(get_value)
        return s_node

    if str(parsed_network[0]) is not "?":
        s_node = MultiNode(parsed_network[0], [])

        for child in parsed_network[1:][0]:
            get_value = rule_to_tree_augmented(child, number=number + 1)
            s_node.children.append(get_value)
        return s_node

    return MultiNode(parsed_network, [])


@lru_cache(maxsize=None)
def get_torch_device():
    return torch.device('cuda' if torch.cuda.is_available() else 'cpu')



from pymetheus.pymetheus import LogicNet
import pytest
from pymetheus.utils.exceptions import DobuleInitalizationException
class TestConstantsClass:

    def test_size(self):
        logicnet = LogicNet()
        logicnet.constant("Rome", argument_size=10)

        assert len(logicnet.constants["Rome"]) == 10

    def test_double(self):

        with pytest.raises(DobuleInitalizationException):
            logicnet = LogicNet()
            logicnet.constant("Rome", argument_size=10)
            logicnet.constant("Rome", argument_size=10)






from pyparsing import (alphanums, alphas, delimitedList, Forward, Group, Keyword, Literal, opAssoc, operatorPrecedence,
                       Suppress, Word, dblQuotedString, removeQuotes)


def _parse_formula(text):
    """
    >>> formula = "p(a,b)"
    >>> print(_parse_formula(formula))
    ['p', ['a', 'b']]

    >>> formula = "~p(a,b)"
    >>> print(_parse_formula(formula))
    ['~', ['p', ['a', 'b']]]

    >>> formula = "=(a,b)"
    >>> print(_parse_formula(formula))
    ['=', ['a', 'b']]

    >>> formula = "<(a,b)"
    >>> print(_parse_formula(formula))
    ['<', ['a', 'b']]

    >>> formula = "~p(a)"
    >>> print(_parse_formula(formula))
    ['~', ['p', ['a']]]

    >>> formula = "~p(a)|a(p)"
    >>> print(_parse_formula(formula))
    [['~', ['p', ['a']]], '|', ['a', ['p']]]

    >>> formula = "p(a) | p(b)"
    >>> print(_parse_formula(formula))
    [['p', ['a']], '|', ['p', ['b']]]

    >>> formula = "~p(a) | p(b)"
    >>> print(_parse_formula(formula))
    [['~', ['p', ['a']]], '|', ['p', ['b']]]

    >>> formula = "p(f(a)) | p(b)"
    >>> print(_parse_formula(formula))
    [['p', [['f', ['a']]]], '|', ['p', ['b']]]

    >>> formula = "p(a) | p(b) | p(c)"
    >>> print(_parse_formula(formula))
    [['p', ['a']], '|', ['p', ['b']], '|', ['p', ['c']]]

    >>> formula = 'p("http://dbpedia.org/ontology/MeanOfTransportation_,_Instrument","1b")'
    >>> print(_parse_formula(formula))
    ['p', ['http://dbpedia.org/ontology/MeanOfTransportation_,_Instrument', '1b']]

    >>> formula = '"http://dbpedia.org/ontology/range"("http://dbpedia.org/ontology/MeanOfTransportation_,_Instrument","1b")'
    >>> print(_parse_formula(formula))
    ['http://dbpedia.org/ontology/range', ['http://dbpedia.org/ontology/MeanOfTransportation_,_Instrument', '1b']]

    """
    left_parenthesis, right_parenthesis, colon = map(Suppress, "():")
    exists = Keyword("exists")
    forall = Keyword("forall")
    implies = Literal("->")
    or_ = Literal("|")
    and_ = Literal("&")
    not_ = Literal("~")
    equiv_ = Literal("%")

    symbol = Word(alphas + "_" + "?" + ".", alphanums + "_" + "?" + "." + "-") | \
        dblQuotedString.setParseAction(removeQuotes)

    term = Forward()
    term << (Group(symbol + Group(left_parenthesis +
                                  delimitedList(term) + right_parenthesis)) | symbol)

    pred_symbol = Word(alphas + "_" + ".", alphanums + "_" + "." + "-") | Literal("=") | Literal("<") | \
        dblQuotedString.setParseAction(removeQuotes)
    literal = Forward()
    literal << (Group(pred_symbol + Group(left_parenthesis + delimitedList(term) + right_parenthesis)) |
                Group(not_ + pred_symbol + Group(left_parenthesis + delimitedList(term) + right_parenthesis)))

    formula = Forward()
    forall_expression = Group(forall + delimitedList(symbol) + colon + formula)
    exists_expression = Group(exists + delimitedList(symbol) + colon + formula)
    operand = forall_expression | exists_expression | literal

    formula << operatorPrecedence(operand, [(not_, 1, opAssoc.RIGHT),
                                            (and_, 2, opAssoc.LEFT),
                                            (or_, 2, opAssoc.LEFT),
                                            (equiv_, 2, opAssoc.RIGHT),
                                            (implies, 2, opAssoc.RIGHT)])
    result = formula.parseString(text, parseAll=True)

    return result.asList()[0]



# -*- coding: utf-8 -*-
"""Main module."""
import itertools
import random

import numpy as np
import tqdm

from pymetheus.logics import LogicEnum, Logic
from pymetheus.logics.fuzzy_logic import *
from pymetheus.parser import rule_parser as parser
from pymetheus.utils.exceptions import DobuleInitalizationException
from pymetheus.utils.functionalities import batching


class LogicNet:
    def __init__(self, universal_aggregator=lambda x: torch.mean(x),
                 differentiable_logic: Logic = LogicEnum.default().value):

        self.device = get_torch_device()

        self.constants = {}
        self.networks = {"->": differentiable_logic.residual,
                         "&": differentiable_logic.t_norm,
                         "~": differentiable_logic.negation,
                         "|": differentiable_logic.t_conorm,
                         "%": differentiable_logic.equal}
        self.rules = {}
        self.axioms = {}
        self.variables = {}
        self.universal_aggregator = universal_aggregator

    def variable(self, label, domain, labelled=True):
        if labelled:
            self.variables[label] = list(map(lambda x: self.constants[x], domain))
        else:
            self.variables[label] = list(map(lambda x: torch.tensor(x), domain))

    def predicate(self, predicate, network=None, arity=2, argument_size=5, overwrite=False):
        """
        Creates a Neural Network for a string symbol that identifies a predicate
        :param predicate:
        :param network:
        :param arity:
        :param argument_size:
        :param overwrite:
        :return:
        """
        if predicate in self.networks.keys() and overwrite is False:
            raise DobuleInitalizationException("Overwrite behaviour is off, error on double declaration of predicate.",
                                               predicate)

        if network:
            self.networks[predicate] = network
        else:
            self.networks[predicate] = Predicate(argument_size * arity).to(self.device)
            # self.networks[predicate] = LinearPredicate(argument_size * arity).to(self.device)
        self.axioms[predicate] = []  # initializes list of training samples

    def constant(self, name, definition=None, argument_size=5, optimize=False, overwrite=False):
        """
        Creates a (logical) constant in the model. The representation for the constant can be given or learned
        :param name: name of the constant
        :param definition: numpy definition of the variable
        :param argument_size: size of the constant
        :param optimize:
        :param overwrite:
        :return:
        """
        if name in self.constants and overwrite is False:
            raise DobuleInitalizationException("Overwrite behaviour is off, error on double declaration of constant.",
                                               name)

        if definition is None:
            self.constants[name] = torch.randn(argument_size, device=self.device, requires_grad=True).float()
        else:
            self.constants[name] = torch.tensor(definition, device=self.device, requires_grad=optimize).float()

    def universal_rule(self, rule):
        """
        Adds a universally quantified rule to the KB
        :param rule:
        :return:
        """
        parsed_rule = (parser._parse_formula(rule))
        parsed_rule_axiom = parsed_rule[-1]  # remove variables
        vars = parsed_rule[1:-1]  # get variables

        tree_rule = rule_to_tree_augmented(parsed_rule_axiom)
        net = QuantifiedFormula(tree_rule, self.networks, vars, self.universal_aggregator)
        self.rules[rule] = net

    def function(self, name, in_size, out_size):
        """
        Defines a function R^n -> R^m
        :param name:
        :param in_size:
        :param out_size:
        :return:
        """
        self.networks[name] = Function(in_size, out_size)

    def _add_true_axiom(self, predicate, constants):
        """
        Helper method to add a new axiom to the LogicNet
        Useful for bulk initialization
        :param predicate:
        :param constants:
        :return:
        """
        self.axioms[predicate].append((constants, 1))

    def _add_false_axiom(self, predicate, constants):
        """
        Helper method to add a new negative axiom to the LogicNet
        Useful for bulk initialization
        :param predicate:
        :param constants:
        :return:
        """
        self.axioms[predicate].append((constants, 0))

    def knowledge(self, fact):
        """
        Adds a knowledge fact into the training set
        :param fact:
        :return:
        """
        parsed_formula = parser._parse_formula(fact)

        if parsed_formula[0] == "~":  # negative axiom
            self._add_false_axiom(parsed_formula[1][0], parsed_formula[1][1])
        else:
            self._add_true_axiom(parsed_formula[0], parsed_formula[1])

    def compute_grounded_axiom(self, axiom):
        model = self.networks[axiom]  # get the predicate network

        training_examples = self.axioms[axiom]  # get the training samples related to the eaxiom
        targets = []
        inputs = []

        for training_example in training_examples:
            local = []
            arguments_of_predicate = training_example[0]

            for a in arguments_of_predicate:
                local.append(self.constants[a])

            fuzzy_value_to_predict = training_example[1]

            target = torch.from_numpy(np.array([fuzzy_value_to_predict])).type(torch.float)
            targets.append(target)
            inputs.append(local)

        targets = list(targets)
        inputs = list(inputs)

        targets = torch.stack(targets).to(self.device)
        send = list(zip(*inputs))
        send = map(list, send)

        computed = model(send)
        outputs = torch.abs(computed - targets)

        return outputs

    def compute_quantified_rule(self, r_model, batch_size):
        rule_accumulator = []
        temp = {}
        for k in r_model.vars:
            temp[k] = self.variables[k]

        # temp = {k: v for k, v in filter(lambda t: t[0] in r_model.vars, self.variables.items())}

        # {?a : v_Rome, v_Paris, ?b : v_Italy, v_France}
        # forall ?a,?b K(?a,?b) -> P(?b,?a)
        prepare_iterator = (itertools.product(*temp.values()))
        for var_inputs in batching(batch_size, prepare_iterator):
            variables = list(var_inputs)
            inputs = {}

            # [[9,5], [1,2]]

            for index, var in enumerate(r_model.vars):
                inputs[var] = list(map(lambda x: x[index], variables))

            # {a : [9, 1], b: [5, 2]}

            output = r_model(inputs)

            rule_accumulator.append(output.reshape(-1))

        return 1 - self.universal_aggregator(torch.cat(rule_accumulator))

    def define_optimizer(self, learning_rate):
        parameters = set()

        for const in self.constants:
            parameters |= {self.constants[const]}

        for network in self.networks:
            try:
                parameters |= set(self.networks[network].parameters())
            except Exception as e:
                print(e)
                pass
        return torch.optim.Adam(parameters, lr=learning_rate)
        # return adabound.AdaBound(parameters, lr=1e-3, final_lr=0.1)

    def fit(self, epochs=100, grouping=36, learning_rate=.01):

        optimizer = self.define_optimizer(learning_rate)
        pbar = tqdm.tqdm(total=epochs)

        for epoch in range(0, epochs):
            optimizer.zero_grad()
            to_be_optimized = []
            check_satisfiability = []

            to_train = list(self.axioms.keys()) + list(self.rules.keys())

            random.shuffle(to_train)

            for a in self.variables:
                random.shuffle(self.variables[a])

            for rule_axiom in to_train:
                if "forall" in rule_axiom:
                    r_model = self.rules[rule_axiom]
                    output = self.compute_quantified_rule(r_model, grouping)

                    to_be_optimized.append(output)
                    # output.backward()
                    # optimizer.step()
                    # optimizer.zero_grad()
                    # to_be_optimized.append(output)
                    check_satisfiability.append(output.item())
                else:
                    axiom = rule_axiom
                    training_examples = self.axioms[axiom]  # get the training samples related to the axiom

                    if not training_examples:
                        continue

                    truth_values = self.compute_grounded_axiom(axiom)
                    mean_truth_value = torch.mean(truth_values)
                    # to_be_optimized.append(mean_truth_value)
                    squeezed = truth_values.squeeze()  # torch.mean(truth_values)

                    # mean_truth_value.backward()
                    # optimizer.step()
                    # optimizer.zero_grad()
                    if squeezed.dim() > 0:
                        to_be_optimized.extend(squeezed)
                    else:
                        to_be_optimized.append(squeezed)

                    check_satisfiability.append(mean_truth_value.item())

            output = torch.mean(torch.stack(to_be_optimized))

            output.backward()
            optimizer.step()

            with torch.no_grad():
                current_sat = 1 - np.mean(check_satisfiability)

            pbar.set_description("Current Satisfiability %f)" % current_sat)

            pbar.update(1)

            if current_sat > 0.99:
                break

    def reason(self, formula, verbose=False, grouping=32):
        with torch.no_grad():
            parsed_formula = parser._parse_formula(formula)
            if "forall" in formula:
                parsed_rule = (parser._parse_formula(formula))
                parsed_rule_axiom = parsed_rule[-1]  # remove variables
                vars = parsed_rule[1:-1]  # get variables
                tree_rule = rule_to_tree_augmented(parsed_rule_axiom)
                r_model = QuantifiedFormula(tree_rule, self.networks, vars, self.universal_aggregator)
                rule_accumulator = []
                temp = {}
                for k in r_model.vars:
                    temp[k] = self.variables[k]
                # temp = {k: v for k, v in filter(lambda t: t[0] in r_model.vars, self.variables.items())}
                prepare_iterator = (itertools.product(*temp.values()))
                for var_inputs in batching(grouping, prepare_iterator):
                    variables = list(var_inputs)
                    inputs = {}
                    for index, var in enumerate(r_model.vars):
                        inputs[var] = list(map(lambda x: x[index], variables))
                    output = r_model(inputs)
                    rule_accumulator.append(output.reshape(-1))
                value = self.universal_aggregator(torch.cat(rule_accumulator)).cpu().detach().numpy()

                return value
            else:
                if parsed_formula[0] == "~":
                    predicate = parsed_formula[1][0]
                    data = parsed_formula[1][1]
                else:
                    predicate = parsed_formula[0]
                    data = parsed_formula[1]
                model = self.networks[predicate]

                current_input = []
                for element in data:
                    current_input.append(self.constants[element].to(self.device).reshape(1, -1))

                val = model(current_input).cpu().detach().numpy()[0][0]
                if parsed_formula[0] == "~":
                    if verbose:
                        print(formula + ": " + str(1 - val), end="\n")
                    return 1 - val
                else:
                    if verbose:
                        print(formula + ": " + str(val), end="\n")
                    return val



from pymetheus.pymetheus import LogicNet
import pytest
from pymetheus.utils.exceptions import DobuleInitalizationException
class TestPredicatesClass:

    def test_double(self):

        with pytest.raises(DobuleInitalizationException):
            logicnet = LogicNet()
            logicnet.predicate("country", argument_size=10)
            logicnet.predicate("country", argument_size=10)





import numpy as np
import torch
from torch import nn

from pymetheus.logics.interfaces import Logic
from pymetheus.utils.functionalities import get_torch_device


class LukasiewiczLogic(Logic):
    """
    Implements the Fuzzy logic defined by Lukasiewicz.
    """
    class Negation(nn.Module):
        def __init__(self):
            super(LukasiewiczLogic.Negation, self).__init__()

        def forward(self, x):
            return 1 - x

    class TNorm(nn.Module):
        def __init__(self):
            super(LukasiewiczLogic.TNorm, self).__init__()

        def forward(self, x, y):
            assert x.shape == y.shape
            baseline = torch.from_numpy(np.array([0])).type(torch.FloatTensor).to(get_torch_device())

            val = x + y - 1

            return torch.max(baseline, val)

    class TConorm(nn.Module):
        def __init__(self):
            super(LukasiewiczLogic.TConorm, self).__init__()

        def forward(self, x, y):
            assert x.shape == y.shape
            baseline = torch.from_numpy(np.array([1])).type(torch.FloatTensor).to(get_torch_device())
            return torch.max(baseline, x + y)

    class Equal(nn.Module):
        def __init__(self):
            super(LukasiewiczLogic.Equal, self).__init__()

        def forward(self, x, y):
            assert x.shape == y.shape
            return (1 - torch.abs(x - y))

    class Residual(nn.Module):
        # Luk
        def __init__(self):
            super(LukasiewiczLogic.Residual, self).__init__()

        def forward(self, x, y):
            assert x.shape == y.shape
            baseline = torch.from_numpy(np.array([1])).type(torch.FloatTensor).to(get_torch_device())
            return torch.min(baseline, 1 - x + y)
